"""
Parsers
=======

Parsers transform the modules defined by a user into an object usable by
the workers to actually parse a web page.
"""

import os
import re
import importlib
from urllib.parse import urlparse
from typing import Tuple, Dict, List, Callable

from bs4 import BeautifulSoup

from .decorators import DATA_NAME, FOLLOW_NAME, EXCLUDE_NAME, LINKS_NAME


class Url(str):
    """
    An enhanced string object to simplify handling URLs
    """

    @property
    def netloc(self):
        """
        The netloc (more commonly called the domain) of the url
        """

        return urlparse(self).netloc

    @property
    def domain(self):
        """
        Alias for netloc
        """

        return self.netloc

    @property
    def scheme(self):
        """
        The scheme of the url, for example `http` or `https`
        """

        return urlparse(self).scheme

    @property
    def path(self):
        """
        The path of the url, excluding any anchors or query params
        """

        return urlparse(self).path

    @property
    def query(self):
        """
        The query string on the url
        """

        return urlparse(self).query

    def match(self, search):
        return re.search(search, self) is not None


class BeautifulSoupStar(BeautifulSoup):
    """
    An enhanced BeautifulSoup object, with additional helper methods
    """

    def __init__(self, response):
        self.response = response
        self.request = self.response.request
        self.status_code = response.status_code
        self.url = Url(self.request.url)
        super().__init__(markup=self.response.text, features='html.parser')

    def find_text(self, text):
        result = self.find(text=re.compile(text))
        if result:
            return result.parent

    def find_class(self, klass):
        regex = re.compile(klass)
        return self.find_all(attrs={'class': regex})

    def harmonize(self, url):
        url = Url(url)
        # If the link is relative, use the domain from the original request
        # Actually I think we can just use the netloc without any formatting
        netloc = url.netloc or self.url.netloc

        # Remove starting slashes from path, they'll be added later
        if url.path.startswith('/'):
            path = url.path[1:]
        else:
            path = url.path

        if url.query:
            query = f'?{url.query}'
        else:
            query = ''

        return Url(f"{self.url.scheme}://{netloc}/{path}{query}")

    def links(self):
        return [self.harmonize(a['href']) for a in self.find_all(href=True)]

    def title(self):
        return self.find('title').text


class Parser(object):
    """
    Responsible for applying a user's parser module to the response from a
    loader.
    """

    @classmethod
    def from_string(klass, string):
        string = string.replace('.py', '').replace(os.path.sep, '.')
        module = importlib.import_module(string)
        return klass(module)

    def __init__(self, module):
        self.module = module

    def _get_soupstars_attr(self, name: str, multiple: bool = False):
        results = []
        for k, v in vars(self.module).items():
            if hasattr(v, '_soupstars') and v._soupstars == name:
                attr = v
                if multiple:
                    results.append(v)
                else:
                    return v
        else:
            return results

    def parsers(self) -> Dict:
        """
        A dictionary containing
        """

        parsers = list(self._get_soupstars_attr(DATA_NAME, multiple=True))
        return {p.__name__: p for p in parsers}

    def seeds(self) -> List:
        """
        A list of urls defined by the parser, used as the starting points when
        running a crawler.
        """

        seeds = getattr(self.module, 'seeds', [])
        seeds.extend(getattr(self.module, 'urls', []))
        if hasattr(self.module, 'url'):
            seeds.append(self.module.url)
        return [Url(s) for s in seeds]

    def exclude(self, soup: BeautifulSoupStar) -> bool:
        """
        A function pulled from the module to determine if the page's data
        should be excluded from the run
        """

        excludor = self._get_soupstars_attr(EXCLUDE_NAME)
        if not excludor:
            excludor = lambda x: False
        return excludor(soup)

    def follow(self, url: Url) -> bool:
        """
        A function pulled from the module to determine if a particular link
        should be followed (ie downloaded)
        """

        validator = self._get_soupstars_attr(FOLLOW_NAME)
        if not validator:
            validator = lambda x: True
        return validator(url)

    def data(self, soup: BeautifulSoupStar) -> Dict:
        """
        Loads the data and any associated errors from applying the user's
        parsing function.
        """

        data = {}
        errors = {}
        for k, v in self.parsers().items():
            try:
                data[k] = v(soup)
                errors[k] = None
            except Exception as err:
                data[k] = None
                errors[k] = str(err)
        else:
            return {
                "data": data,
                "errors": errors,
                "url": soup.url,
                "status_code": soup.status_code
            }

    def links(self, soup: BeautifulSoupStar) -> List[Url]:
        """
        Uses the `links` function defined on the parser to pull links from
        the response.
        """

        for k,v in vars(self.module).items():
            if hasattr(v, '_soupstars') and v._soupstars == LINKS_NAME:
                links = v(soup)
                break
        else:
            links = soup.links()

        items = list(filter(self.follow, links))
        return items

    def parse(self, response):
        soup = BeautifulSoupStar(response)
        if self.exclude(soup):
            data = None
        else:
            data = self.data(soup)
        links = self.links(soup)
        return data, links



"""
Soup Stars is a framework for building web parsers with Python. It is designed
to make building, deploying, and scheduling web parsers easier by simplifying
what you need to get started.
"""

from .version import __version__
from .api import run, fetch, parse
from .parsers import BeautifulSoupStar, Url
from .decorators import data, follow, exclude, links



"""
Logger
======

The Logger class provides shortcuts for colorized logs generated by objects
and the CLI.
"""

from datetime import datetime
from click import secho, style


class Logger(object):
    """
    :param name: The name for the parser. Used to identify different objects,
    such as various workers.
    :type name: str
    """

    format_str = "[{level}] {ts} {name} - {msg}"
    INFO = 'cyan'
    FAIL = 'red'
    OKAY = 'green'
    WARN = 'yellow'

    def __init__(self, name):
        self.name = str(name)

    def log(self, level, message):
        ts = str(datetime.now())[11:-4]
        name = style(self.name, bold=True)
        res = self.format_str.format(
            level=level,
            msg=message,
            name=name,
            ts=ts
        )
        secho(res)

    def info(self, message):
        level = style('INFO', bold=True, fg=self.INFO)
        self.log(level, message)

    def okay(self, message):
        level = style('OKAY', bold=True, fg=self.OKAY)
        self.log(level, message)

    def fail(self, message):
        level = style('FAIL', bold=True, fg=self.FAIL)
        self.log(level, message)

    def warn(self, message):
        level = style('WARN', bold=True, fg=self.WARN)
        self.log(level, message)



"""
Writers control where the results of crawls are stored.
"""

import csv
import importlib
import json
import os

from .resources import ResultsResource
from .config import Config


class Writer(object):

    @classmethod
    def from_string(klass, string, url=None):
        class_name = f"{string.capitalize()}Writer"
        module = importlib.import_module('soupstars.writers')
        _Writer = getattr(module, class_name)
        return _Writer(url=url)

    def __init__(self, url=None):
        self.url = url


# I think we could do a TextWriter that optionally logs to the current direcfory
# or the soupstars home directory
class JsonWriter(Writer):
    def write(self, result):
        with open(self.url, 'a') as o:
            json.dump(result, o)
            o.write('\n')


class CsvWriter(Writer):
    def flatten(self, result):
        _result = {}
        errors = {f'{k}_error': v for k, v in result['errors'].items()}
        data = {k: str(v) for k, v in result['data'].items()}
        _result.update(url=result['url'])
        _result.update(status_code=result['status_code'])
        _result.update(data)
        _result.update(errors)
        return _result

    def write(self, result):
        result = self.flatten(result)
        with open(self.url, 'a') as o:
            _writer = csv.DictWriter(o, fieldnames=result.keys(), delimiter='\t')
            if o.tell() == 0:
                _writer.writeheader()
            _writer.writerow(result)


class CloudWriter(Writer):

    def __init__(self, run_id, parser_id, host=None, token=None):
        self.parser_id = parser_id
        self.run_id = run_id
        self.config = Config(host=host, token=token)
        self.resource = ResultsResource(config=self.config)

    def write(self, result):
        return self.resource.post(
            run_id=self.run_id,
            parser_id=self.parser_id,
            data=result['data'],
            url=result['url'],
            status_code=result['status_code'],
            errors=result['errors']
        )



import requests

from .config import Config


class Resource(object):

    def __init__(self, config):
        self.config = config

    def send(self, path, method, **kwargs):
        url = self.config.host + self.route
        headers = {'Authorization': f'jwt {self.config.token}'}
        resp = requests.request(
            url=url,
            method=method,
            headers=headers,
            json=kwargs
        )
        return resp

    def get(self, **kwargs):
        return self.send(self.route, method='GET', **kwargs)

    def post(self, **kwargs):
        return self.send(self.route, method='POST', **kwargs)

    def patch(self, **kwargs):
        return self.send(self.route, method='PATCH', **kwargs)


class AuthResource(Resource):
    route = '/auth'

    def post(self, email, password):
        return super().post(email=email, password=password)


class ConfigResource(Resource):
    route = '/config'


class StatusResource(Resource):
    route = '/status'


class UsersResource(Resource):
    route = '/users'

    def post(self, email, password):
        return super().post(email=email, password=password)


class AccountsResource(Resource):
    route = '/accounts'


class ProfileResource(Resource):
    route = '/profile'


class PlansResource(Resource):
    route = '/plans'


class ParsersResource(Resource):
    route = '/parsers'


class RunsResource(Resource):
    route = '/runs'

    def post(self, parser_id):
        return super().post(parser_id=parser_id)


class ResultsResource(Resource):
    route = '/results'

    def post(self, run_id, parser_id, status_code, url, data, errors):
        return super().post(run_id=run_id, parser_id=parser_id,
                         status_code=status_code, url=url, data=data,
                         errors=errors)


class InstancesResource(Resource):
    route = '/instances'



"""
API
===

Entrypoints for some of the functionality exposed by the package
"""

from .runners import Runner
from .loaders import Loader
from .parsers import BeautifulSoupStar, Url


def run(parser, **kwargs):
    """
    Executes a parser, returning the data from the run
    """

    return Runner(parser, **kwargs).run()


def fetch(url) -> BeautifulSoupStar:
    """
    Downloads a url and converts it into a BeautifulSoupStars object
    """

    loader = Loader()
    load = loader.load(url)
    soup = BeautifulSoupStar(load)
    return soup


def parse(parser, url):
    """
    Download a url and process it with a parser, returning the data and links
    discovered by the parsers
    """

    loader = Loader()
    load = loader.load(url)
    data, links = parser.parse(load)
    return data, links



import click
import os
import sys
import textwrap
import getpass

from .printers import jsonify, pythonify
from ..config import Config
from ..resources import UsersResource, StatusResource, AuthResource, ParsersResource, ProfileResource


@click.group()
@click.option('--token', '-t', help="Token to use. Default None")
@click.option('--host', '-h', default=None, help="Host to use")
@click.pass_context
def cloud(context, token, host):
    """
    Commands to interact with SoupStars cloud.
    """

    context.obj = Config(token=token, host=host)


@cloud.command()
@click.pass_obj
def health(config):
    """
    Print the status of the SoupStars api
    """

    status = StatusResource(config=config)
    resp = status.get()
    jsonify(resp.json())


@cloud.command()
@click.pass_obj
def register(config):
    """
    Register a new account on SoupStars cloud
    """

    email = input('Email: ')
    password = getpass.getpass(prompt='Password: ')
    password2 = getpass.getpass(prompt='Confirm password: ')

    if password != password2:
        print("Passwords did not match.")
        return

    users = UsersResource(config=config)
    resp = users.post(email=email, password=password)

    if resp.ok:
        auth = AuthResource(config=config)
        resp = auth.post(email=email, password=password)
        config.token = resp.json()['access_token']
        config.save()
    jsonify(resp.json())


@cloud.command()
@click.pass_obj
def login(config):
    """
    Log in with an existing email
    """

    email = input('Email: ')
    password = getpass.getpass(prompt='Password: ')
    auth = AuthResource(config=config)
    resp = auth.post(email=email, password=password)
    if resp.ok:
        config.token = resp.json()['access_token']
        config.save()

    jsonify(resp.json())


@cloud.command()
@click.pass_obj
def whoami(config):
    """
    Print the email address of the current user
    """

    profile = ProfileResource(config=config)
    resp = profile.get()
    if resp.ok:
        jsonify(resp.json().get('email'))
    else:
        jsonify(resp.json())


@cloud.command()
@click.pass_obj
def ls(config):
    """
    Show the parsers uploaded to SoupStars cloud
    """

    parsers = ParsersResource(config=config)
    resp = parsers.get()
    jsonify(resp.json())


# TODO
# @cloud.command()
@click.option('--module', '-m', required=True, help="Name of the parser to push")
@click.pass_obj
def push(client, module):
    """
    Push a parser to SoupStars cloud
    """

    parsers = ParsersResource(config=config)
    jsonify(resp.json())


# TODO: update the api to allow pulling a parse by name
# @cloud.command()
@click.option('--module', '-m', required=True, help="Name of the parser to pull")
@click.pass_obj
def pull(client, module):
    """
    Pull a parser from SoupStars cloud into a local module
    """

    resp = client.pull(module)
    data = resp.json()
    with open(data['parser']['name'], 'w') as o:
        o.write(data['module'])
    jsonify({"state": "done", "response": data})


# TODO
# @cloud.command()
@click.option('--module', '-m', required=True, help="Name of the parser to create")
@click.pass_obj
def run(client, module):
    """
    Run a parser on SoupStars cloud
    """

    resp = client.run(module)
    jsonify(resp.json())


# TODO
# @cloud.command()
@click.option('--module', '-m', required=True, help="Name of the parser to show")
@click.option('--json/--no-json', default=False, help="Show parser details in JSON")
@click.pass_obj
def show(client, module, json):
    """
    Show the contents of a parser on SoupStars cloud
    """

    resp = client.pull(module)

    if json:
        jsonify(resp.json())
    else:
        pythonify(resp.json()['module'])


# TODO
# @cloud.command()
@click.option('--module', '-m', required=True, help="Name of the parser to test")
@click.pass_obj
def results(client, module):
    """
    Print results of a parser
    """

    resp = client.results(module)
    jsonify(resp.json())
